// UselessMine.cpp : Defines the entry point for the DLL application.
//

#include "bzfsAPI.h"
#include <vector>
#include <string>
#include <sstream>
#include <cmath>
#include <algorithm>

#include <boost/array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>

#define kMineShotID 100
#define kMineShotIDMax 142

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


BZ_GET_PLUGIN_VERSION



class UMMine;

typedef boost::shared_ptr<UMMine> UMMinePtr;
typedef std::vector<UMMinePtr>::iterator ml_iterator;

//A list of all the mines currently on the field
std::vector<UMMinePtr> mineList;

//Used to keep track of the killer.  Not a good solution.
int lastKillerID = -1;

//I just wanted to use this :D
//It's definately overkill for what I'm using it for.
//The exponent must be an integer constant, known at compile time.
namespace dcd {
  namespace math {
    template <unsigned int N, typename T> class pow {
    public:
      static T algorithm(const T &x)
    {
        return x * pow<N-1, T>::algorithm(x);
    }
    };

    template <typename T> class pow<1, T> {
    public:
      static T algorithm(const T &x)
    {
        return x;
    }
    };
  }
}

template <unsigned int N, typename T>
T pow_m(T base)
{
  return dcd::math::pow<N, T>::algorithm(base);
}



class UMMine {
public:
  //The player who left the mine
  int playerID;
  //Does this mine exist
  bool exists;
  //The position of the mine.
  boost::array<float, 3> pos;
  //Controls the blast radius.
  float mineTimeDuration;

  UMMine();
  UMMine(int id, const boost::array<float, 3>& position);
  virtual ~UMMine() { }

  //Blows up the mine
  virtual void fire() = 0;
};

inline
UMMine::UMMine() : playerID(-1), exists(true), mineTimeDuration(0)
{
  pos.assign(0);
}

inline
UMMine::UMMine(int id, const boost::array<float, 3>& position)
: playerID(id), exists(true), pos(position), mineTimeDuration(0)
{ }



// Fires a shock wave.
class ShockWaveMine : public UMMine {
public:
  ShockWaveMine() {}
  ShockWaveMine(int id, const boost::array<float, 3>& position);
  virtual void fire();
};

inline
ShockWaveMine::ShockWaveMine(int id, const boost::array<float, 3>& position)
: UMMine(id, position)
{ }

void ShockWaveMine::fire()
{
  //the mine does not exist anymore
  exists = false;
  bz_fireWorldWep("SW", 2.0f, BZ_SERVER, pos.c_array(), 0, 0, kMineShotID+26, 0.0f);
}



// Spherical: Fires in a spherical pattern.
// Reg: Fires regular bullets.
// Mine: It's a mine, of course.
class SphericalRegMine : public UMMine {
public:
  SphericalRegMine() {}
  SphericalRegMine(int id, const boost::array<float, 3>& position);
  virtual void fire();
};

inline
SphericalRegMine::SphericalRegMine(int id, const boost::array<float, 3>& position)
: UMMine(id, position)
{
  mineTimeDuration = static_cast<float>(bz_getBZDBDouble("_shockOutRadius") /
    (bz_getBZDBDouble("_rFireAdVel") * bz_getBZDBDouble("_shotSpeed")) * 4);
}

void SphericalRegMine::fire()
{
  exists = false;
  float dt = 0.25f;

  //straight up
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/2, 0, kMineShotID, dt);

  //straight down
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 3*M_PI/2, 0, kMineShotID+1, dt);

  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 0, kMineShotID+2, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI/4, kMineShotID+3, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI/2, kMineShotID+4, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 3*M_PI/4, kMineShotID+5, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI, kMineShotID+6, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 5*M_PI/4, kMineShotID+7, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 3*M_PI/2, kMineShotID+8, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 7*M_PI/4, kMineShotID+9, dt);

  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, 0, kMineShotID+10, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, M_PI/4, kMineShotID+11, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, M_PI/2, kMineShotID+12, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, 3*M_PI/4, kMineShotID+13, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, M_PI, kMineShotID+14, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, 5*M_PI/4, kMineShotID+15, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, 3*M_PI/2, kMineShotID+16, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), M_PI/4, 7*M_PI/4, kMineShotID+17, dt);

  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, 0, kMineShotID+18, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, M_PI/4, kMineShotID+19, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, M_PI/2, kMineShotID+20, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, 3*M_PI/4, kMineShotID+21, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, M_PI, kMineShotID+22, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, 5*M_PI/4, kMineShotID+23, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, 3*M_PI/2, kMineShotID+24, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 7*M_PI/4, 7*M_PI/4, kMineShotID+25, dt);

  //That's alot of world weapons!
  //Try dodging this, GU players! ;)
}


// Cylindrical: Fires in a cylindrical pattern.
// Reg: Fires regular bullets.
// Mine: It's a mine, of course.
class CylindricalRegMine : public UMMine {
public:
  CylindricalRegMine() {}
  CylindricalRegMine(int id, const boost::array<float, 3>& position);
  virtual void fire();
};

inline
CylindricalRegMine::CylindricalRegMine(int id, const boost::array<float, 3>& position)
: UMMine(id, position)
{
  mineTimeDuration = static_cast<float>(bz_getBZDBDouble("_shockOutRadius") /
    (bz_getBZDBDouble("_rFireAdVel") * bz_getBZDBDouble("_shotSpeed")) * 4);
}

void CylindricalRegMine::fire()
{
  exists = false;
  float dt = 0.25f;

  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 0, kMineShotID+27, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI/4, kMineShotID+28, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI/2, kMineShotID+29, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 3*M_PI/4, kMineShotID+30, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI, kMineShotID+31, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 5*M_PI/4, kMineShotID+32, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 3*M_PI/2, kMineShotID+33, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 7*M_PI/4, kMineShotID+34, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI/6, kMineShotID+35, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, M_PI/3, kMineShotID+36, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 2*M_PI/3, kMineShotID+37, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 5*M_PI/6, kMineShotID+38, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 7*M_PI/6, kMineShotID+39, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 4*M_PI/3, kMineShotID+40, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 5*M_PI/3, kMineShotID+41, dt);
  bz_fireWorldWep("F", mineTimeDuration, BZ_SERVER, pos.c_array(), 0, 11*M_PI/6, kMineShotID+42, dt);
}


bool distanceLess(const boost::array<float, 3>& pos1,
                  const boost::array<float, 3>& pos2,
                  float dist)
{
  return std::sqrt(pow_m<2>(pos1[0] - pos2[0]) +
                   pow_m<2>(pos1[1] - pos2[1]) +
                   pow_m<2>(pos1[2] - pos2[2])) < dist;
}


bool distanceLess(float pos1[3], float pos2[3], float dist)
{
  return std::sqrt(pow_m<2>(pos1[0] - pos2[0]) +
                   pow_m<2>(pos1[1] - pos2[1]) +
                   pow_m<2>(pos1[2] - pos2[2])) < dist;
}

void cleanupMineList()
{
  ml_iterator begin = mineList.begin();
  ml_iterator end = mineList.end();

  while(begin != end) {
    UMMinePtr mine = *begin;
    if(!mine->exists) {
      mineList.erase(begin);
      end = mineList.end();
    }

    if(begin != end)
      ++begin;
  }
}

void sendMineDeathMessage(int victimID, int killerID)
{
  bz_PlayerRecord *killer = bz_getPlayerByIndex(killerID);
  if (!killer) return;

  bz_PlayerRecord *victim = bz_getPlayerByIndex(victimID);
  if (!victim) {
    bz_freePlayerRecord(killer);
    return;
  }

  std::ostringstream victim_msg, all_msg;
  victim_msg << "You were destroyed by " << killer->callsign.c_str() << "'s mine.";
  all_msg << victim->callsign.c_str() << " was destroyed by " << killer->callsign.c_str()
    << "'s mine. " << '[' << mineList.size() << ']';

  std::string pmsg(victim_msg.str());
  std::string amsg(all_msg.str());

  bz_sendTextMessage(BZ_SERVER, victimID, pmsg.c_str());
  bz_sendTextMessage(BZ_SERVER, BZ_ALLUSERS, amsg.c_str());

  bz_freePlayerRecord(killer);
  bz_freePlayerRecord(victim);
}



//This event checks how close a player is to an existing mine, and
//fires a world weapon shock wave if he or she is withing range.
class UMPlayerUpdateHandler : public bz_EventHandler {
public:
  virtual void process(bz_EventData *eventData)
  {
    if(eventData->eventType != bz_ePlayerUpdateEvent)
      return;

    bz_PlayerUpdateEventData *eData = dynamic_cast<bz_PlayerUpdateEventData*>(eventData);
    float dist = static_cast<float>(bz_getBZDBDouble("_shockOutRadius"));

    int i = 0;
    int size = mineList.size();
    while(i < size) {
      UMMinePtr mine = mineList[i];

      //Skip the player who set the mine
      if(eData->playerID == mine->playerID) {
        ++i;
        continue;
      }

      bool withinRange = distanceLess(eData->pos, mine->pos.c_array(), dist);
      if(withinRange && mine->exists) {
        mine->fire();
        lastKillerID = mine->playerID;
      }
      ++i;
    }

    cleanupMineList();
  }
};


//Used to create/set mines
class UMMineCommand : public bz_CustomSlashCommandHandler {
public:
  virtual bool handle (int playerID,
                       bzApiString /*command*/,
                       bzApiString /*message*/,
                       bzAPIStringList* params)
  {
    bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);
    if (!player)
      return true;

    bzAPIStringList &args = *params;

    if(args.size() == 0) {
      bz_sendTextMessage(BZ_SERVER, player->playerID,
                         "Usage: /mine <SW | SR | CR>");
      bz_sendTextMessage(BZ_SERVER, player->playerID,
                         "SW: fires a shockwave.");
      bz_sendTextMessage(BZ_SERVER, player->playerID,
                         "SR: fires bullets in a spherical pattern.");
      bz_sendTextMessage(BZ_SERVER, player->playerID,
                         "CR: fires bullets in a cylindrical pattern.");
      bz_freePlayerRecord(player);
      return true;
    }

    if(args.size() > 0) {
      std::string flag(player->currentFlag.c_str());
      if(flag.find("US") == std::string::npos) {
        bz_sendTextMessage(BZ_SERVER, player->playerID,
                           "You do not have the Useless(US) flag");
        bz_freePlayerRecord(player);
        return true;
      }
    }

    int i = 0;
    int size = args.size();

    //What type of mine
    while(i < size) {
      std::string current(args[i].c_str());
      boost::to_upper(current);

      if(current == "SW") {
        boost::array<float, 3> mine_pos;
        std::copy(player->pos, player->pos+3, mine_pos.begin());

        //The Z(up) coordinate is on the bottom of the tank.
        //Need to raise it up for regular bullets.
        mine_pos[2] += static_cast<float>(bz_getBZDBDouble("_tankHeight")/2);
        UMMinePtr mine(new ShockWaveMine(player->playerID, mine_pos));
        mineList.push_back(mine);
        bz_removePlayerFlag(player->playerID);
        break;
      }
      else if(current == "SR") {
        boost::array<float, 3> mine_pos;
        std::copy(player->pos, player->pos+3, mine_pos.begin());

        //The Z(up) coordinate is on the bottom of the tank.
        //Need to raise it up for regular bullets.
        mine_pos[2] += static_cast<float>(bz_getBZDBDouble("_tankHeight")/2);
        UMMinePtr mine(new SphericalRegMine(player->playerID, mine_pos));
        mineList.push_back(mine);
        bz_removePlayerFlag(player->playerID);
        break;
      }
      else if(current == "CR") {
        boost::array<float, 3> mine_pos;
        std::copy(player->pos, player->pos+3, mine_pos.begin());

        //The Z(up) coordinate is on the bottom of the tank.
        //Need to raise it up for regular bullets.
        mine_pos[2] += static_cast<float>(bz_getBZDBDouble("_tankHeight")/2);
        UMMinePtr mine(new CylindricalRegMine(player->playerID, mine_pos));
        mineList.push_back(mine);
        bz_removePlayerFlag(player->playerID);
        break;
      }
      ++i;
    }

    bz_freePlayerRecord(player);

    return true;
  }
};


//Used to make sure that a player's mine is removed if he or she leaves.
class UMJoinPartHandler : public bz_EventHandler {
public:
  virtual void process(bz_EventData *eventData)
  {
    if(eventData->eventType != bz_ePlayerPartEvent)
      return;

    bz_PlayerJoinPartEventData *eData = dynamic_cast<bz_PlayerJoinPartEventData*>(eventData);

    ml_iterator ipos = mineList.begin();
    ml_iterator end = mineList.end();

    while(ipos != end) {
      UMMinePtr mine = *ipos;
      if(eData->playerID == mine->playerID)
        mine->exists = false;
      ++ipos;
    }
  }
};



class UMDieHandler : public bz_EventHandler {
public:
  virtual void process(bz_EventData *eventData)
  {
    if(eventData->eventType != bz_ePlayerDieEvent)
      return;

    bz_PlayerDieEventData *eData = dynamic_cast<bz_PlayerDieEventData*>(eventData);

    //Only interested in those who were killed by a mine
    if((eData->shotID < kMineShotID) || (eData->shotID > kMineShotIDMax))
      return;

    /*
      DON'T USE EDATA->KILLERID, SINCE
      IT IS THE SERVER WHO IS KILLING
     */

    sendMineDeathMessage(eData->playerID, lastKillerID);
  }
};


UMMineCommand mineCommand;
UMJoinPartHandler joinPartHandler;
UMPlayerUpdateHandler playerUpdateHandler;
UMDieHandler dieHandler;

BZF_PLUGIN_CALL int bz_Load(const char* /*commandLine*/)
{
  bz_debugMessage(4,"UselessMine plugin loaded");
  bz_registerCustomSlashCommand ("mine", &mineCommand);
  bz_registerEvent(bz_ePlayerPartEvent, &joinPartHandler);
  bz_registerEvent(bz_ePlayerUpdateEvent, &playerUpdateHandler);
  bz_registerEvent(bz_ePlayerDieEvent, &dieHandler);
  return 0;
}

BZF_PLUGIN_CALL int bz_Unload(void)
{
  bz_debugMessage(4,"UselessMine plugin unloaded");
  bz_removeCustomSlashCommand ("mine");
  bz_removeEvent(bz_ePlayerPartEvent, &joinPartHandler);
  bz_removeEvent(bz_ePlayerUpdateEvent, &playerUpdateHandler);
  bz_removeEvent(bz_ePlayerDieEvent, &dieHandler);
  return 0;
}

// Local Variables: ***
// mode:C++ ***
// tab-width: 8 ***
// c-basic-offset: 2 ***
// indent-tabs-mode: t ***
// End: ***
// ex: shiftwidth=2 tabstop=8
